import numpy as np
import scipy.signal as ss


class SpectralData(object):
    """SpectralData renders power spectral density, spectral moments and
    spectral band estimators, based on input signal time history. Object 
    contains values needed for subsequent fatigue life calculations. 

    Notes
    -----
    A primitive caching is used. Here is how it works: with class
    instantiation, ALL values that are expected to be used are calculated.
    The properties are created and then called. Calculation is done only at
    the beginning and only once.
    """

    def __init__(self, input, dt, window='hanning', nperseg=1280,
                 noverlap=None, psd_trim_length=None):
        """Call parent constructor, analyse input and define fatigue life
        parameters and constants.
        
        
        :param input: [array, string]
                Signal values/file-path. Input can be 1D array of values or
                path to appropriately formated .txt file.
        :param dt: float
                Sampling interval
        :param window: str or tuple or array_like, optional
                Desired window to use. Defaults to ‘hanning’.
        :param nperseg:  int, optional
                Length of each segment. Defaults to 1280.
        :param noverlap: int, optional
                Number of points to overlap between segments. If None, 
                noverlap = nperseg / 2. Defaults to None.

        Properties
        ----------
        data    : array 1D
                Signal values with time spacing dt
        dt      : float
                Time between discreete signal values    
        moments : array 1D
                Spectral moments from m0 to m4    
        psd     : array 2D
                Normalized power spectral density
        t       : float
                Length of signal in seconds, N * dt

        Raises
        ------
        ValueError, for unknown input.
        """        
        if isinstance(input, str):
            self.data = self.readf(input)
        elif isinstance(input, np.ndarray):
            self.data = input
        else:
            raise Exception('Unrecognized Input Error')

        if isinstance(dt, float):
            self.dt = dt
        else:
            raise Exception('Unrecognized Input Error')

        self.t = self.dt * self.data.size

        self.x = np.arange(0, self.t, dt)

        self.trim_length = psd_trim_length

        self.psd = self._calculate_psd(self.data, fs=1.0/self.dt, window=window,
                                       nperseg=nperseg, noverlap=noverlap,
                                       trim=psd_trim_length)

        self.moments = self.calculate_spectral_moments_frequency(self.psd)
        self.moments_omega = self.calculate_spectral_moments_omega(self.psd)
        self._calculate_coefficients()


    def _calculate_psd(self, data, fs=1, window='hamming', nperseg=10280, noverlap=None, trim=None):
        """Calculates PSD using welch estimator."""
        f, p = ss.welch(data, fs=fs, window=window,
                        nperseg=nperseg, noverlap=noverlap)

        psd = np.vstack((f, p)).transpose()

        if trim is None:
            return psd
        else:
            df = fs / nperseg
            trim_idx = int(np.floor(trim * df))

            return psd[:trim_idx, :trim_idx]

    def _calculate_coefficients(self):
        """Calculate all of them."""
        self.m_p = self._get_mp()
        self.m075 = self._get_m075()
        self.m150 = self._get_m150()
        self.nu = self._get_nup()

        self.al075 = self.m075 / (np.sqrt(self.moments[0] * self.m150))
        self.al1 = self.moments[1] / (np.sqrt(self.moments[0] * self.moments[2]))
        self.al2 = self.moments[2] / (np.sqrt(self.moments[0] * self.moments[4]))

    def calculate_spectral_moments_frequency(self, psd):
        """Calculate 0th to 4th PSD moment.
        
        Returns
        -------
        Array [m0,m1,m2,m3,m4].
        """
        f = psd[:, 0]
        p = psd[:, 1]

        m = [
            np.trapz(p, f),
            np.trapz(f*p, f),
            np.trapz(f**2 * p, f),
            np.trapz(f**3 * p, f),
            np.trapz(f**4 * p, f),
        ]

        return np.array(m)
    
    def calculate_spectral_moments_omega(self, psd):
        om = 2 * np.pi * psd[:, 0]
        p = psd[:, 1]
        m = [
            np.trapz(p, om),
            np.trapz(om*p, om),
            np.trapz(om**2 * p, om),
            np.trapz(om**3 * p, om),
            np.trapz(om**4 * p, om),
        ]

        return np.array(m)
    
    def _get_m075(self):
        '''Calculate 0.75th spectral moment.'''
        f = self.psd[:, 0]
        p = self.psd[:, 1]
        
        return np.trapz(f**0.75 * p, f)

    def _get_m150(self):
        '''Calculate 1.5th spectral moment.'''
        f = self.psd[:, 0]
        p = self.psd[:, 1]
        
        return np.trapz(f**1.5 * p, f)
    
    def _get_mp(self):
        '''Calculate M+ / nu_p'''
        return np.sqrt(self.moments[4] / self.moments[2])
    
    def _get_nup(self):
        '''Calculate nu_p - peak intensity for narrowband signal.'''
        return np.sqrt(self.moments[2] / self.moments[0])
    
    def readf(self, filename):
        """Read input file and extract values in form of array (float).
        
        Values inside txt file must be separated with whitespace, e.g. space or newline.
        
        Parameters
        ----------
        filename : string
                   path to file
        
        Returns
        -------
        1D array filled with values. 
        """
        f = open(filename)
        data = np.array(map(float,f.read().split()))
        f.close()
        
        return data


import numpy as np
from scipy.integrate import quad
from .narrowband import Narrowband

class TovoBenasciutti(Narrowband):
    """Class for fatigue life estimation using frequency domain 
    method by Tovo and Benasciutti[1, 2].
      
    References
    ----------
    [1] Roberto Tovo. Cycle distribution and fatigue damage under broadband
        random loading. International Journal of Fatigue, 24(11):1137{
        1147, 2002
    [2] Denis Benasciutti and Roberto Tovo. Spectral methods for lifetime
        prediction under wide-band stationary random processes. International
        Journal of Fatigue, 27(8):867{877, 2005
    [3] Janko Slavič, Matjaž Mršnik, Martin Česnik, Jaka Javh, Miha Boltežar. 
        Vibration Fatigue by Spectral Methods, From Structural Dynamics to Fatigue Damage
        – Theory and Experiments, ISBN: 9780128221907, Elsevier, 1st September 2020
        """
    def __init__(self, spectral_data):
        '''Get needed values from reference object.

        :param spectral_data:  Instance of object SpectralData
        '''     
        Narrowband.__init__(self, spectral_data)

    def _calculate_coefficient(self, method='method 2'):
        """Calculate weigthing parameter b for the Tovo-Benasciutti method. Parameter b is 
            defined by Tovo and Benasciutti [1,2].
        
        :param method:  string
            - 'method 1': `b` weighting parameter `b` is defined by Tovo[1].
            - 'method 2': `b` weighting parameter `b` is defined by Tovo and Benasciutti [2].
                          (This is the improved method)
        :return b: float
        """
        if method == 'method 1': 
            b = self._calculate_coefficient_method_1()
        elif method == 'method 2': 
            b = self._calculate_coefficient_method_2()
        else: 
            raise Exception('Unrecognized Input Error')
        return b


    def _calculate_coefficient_method_1(self):
        """Calculate weigthing parameter b Tovo-Benasciutti method. Parameter b is 
            defined by Tovo[1].
        
        :return b: float
        """
        al1 = self.spectral_data.al1
        al2 = self.spectral_data.al2

        b = min( (al1-al2) / (1.0 - al1), 1.0)
        
        return b

    def _calculate_coefficient_method_2(self):
        """Calculate weigthing parameter b for improved Tovo-Benasciutti method. Parameter b is 
            defined by Tovo and Benasciutti [2].
        
        :return b: float
        """
        al1 = self.spectral_data.al1
        al2 = self.spectral_data.al2

        b = (al1-al2) * (  1.112 * ( 1+ al1*al2 - (al1+al2)  ) * np.exp(2.11*al2) +(al1-al2) ) / ((al2-1)**2)

        return b
        
    def get_PDF(self, s, method='method 2'):
        '''Returns cycle PDF(Probability Density Function) as a function of stress s.

        :param s:  numpy.ndarray
            Stress vector.
        :param method:  string
            - 'method 1': `b` weighting parameter `b` is defined by Tovo[1].
            - 'method 2': `b` weighting parameter `b` is defined by Tovo and Benasciutti [2].
                          (This is the improved method)
        :return pdf: function pdf(s)
        '''
        al2 = self.spectral_data.al2
        m0 = self.spectral_data.moments[0]

        b = self._calculate_coefficient(method=method)

        def pdf(s):
            px = b * ((s / m0) * np.exp( - s**2 / (2 * m0))) + \
                (1 - b) * ((s / (m0 * al2**2)) * np.exp( - s**2 / (2 * al2**2 * m0))) 
            return px
        return pdf(s)

    def get_life(self, C, k, method='method 2', integrate_pdf=False):
        """Calculate fatigue life with parameters C, k, as defined in [3].

        :param C: [int,float]
            Fatigue strength coefficient [MPa**k].
        :param k : [int,float]
            Fatigue strength exponent [/].
        :param method:  string
            - 'method 1': `b` weighting parameter `b` is defined by Tovo[1].
            - 'method 2': `b` weighting parameter `b` is defined by Tovo and Benasciutti [2].
                          (This is the improved method)
        :param integrate_pdf:  boolean
            If true the the fatigue life is estimated by integrating the PDF, 
            Default is false which means that the theoretical equation is used
        :return T: float
            Estimated fatigue life in seconds.
        """ 
        if integrate_pdf:
            d = self.spectral_data.nu / C * \
                quad(lambda s: s**k*self.get_PDF(s, method=method), 
                     a=0, b=np.Inf)[0]
        else:
            m0 = self.spectral_data.moments[0]
            nu = self.spectral_data.nu
            al2 = self.spectral_data.al2
            
            b = self._calculate_coefficient(method=method)

            dNB = self.damage_intesity_NB(m0=m0, nu=nu, C=C, k=k) 
            l = b + ( 1.0 - b ) * al2**(k-1.0)
            d = dNB * l
        
        T = 1.0/d
        return T


import numpy as np
import scipy.special as ss
from scipy.integrate import quad

class Dirlik(object):
    """Class for fatigue life estimation using frequency domain method by Dirlik [1].
   
    References
    ----------
    [1] Turan Dirlik. Application of computers in fatigue analysis. PhD thesis,
        University of Warwick, 1985
    [2] Janko Slavič, Matjaž Mršnik, Martin Česnik, Jaka Javh, Miha Boltežar. 
        Vibration Fatigue by Spectral Methods, From Structural Dynamics to Fatigue Damage
        – Theory and Experiments, ISBN: 9780128221907, Elsevier, 1st September 2020
    """
    def __init__(self, spectral_data):
        '''Get needed values from reference object.

        :param spectral_data:  Instance of object SpectralData
        '''          
        self.spectral_data = spectral_data
        self._calculate_coefficients() 

    def _calculate_coefficients(self):
        '''Calculate coefficients for Dirlik method.
        '''
        m0,m1,m2,_,m4 = self.spectral_data.moments
        
        c = np.empty(8)
        c[0] = ( 1. / np.sqrt(m0) ) 
        c[1] = ( ( m1 / m0 ) * np.sqrt( m2 / m4 ) ) 
        c[2] = ( np.sqrt( 1. / ( m0 * m4 ) ) * m2 ) 
        c[3] = ( 2. * ( c[1] - c[2]**2 ) / ( 1. + c[2]**2 ) ) 
        c[4] = ( ( c[2] - c[1] - c[3]**2 ) / (1 - c[2] - c[3] + c[3]**2 ) ) 
        c[5] = ( ( 1. - c[2] - c[3] + c[3]**2 ) / ( 1. - c[4] ) ) 
        c[6] = ( ( 1. - c[3] - c[5] ) ) 
        c[7] = ( 1.25 * ( c[2] - c[6] - c[5] * c[4] ) / c[3] ) 
        
        self.coeff = c
            

    def get_PDF(self,s):
        '''Returns cycle PDF(Probability Density Function) as a function of stress s.

        :param s:  numpy.ndarray
            Stress vector.
        :return pdf: function pdf(s)
        '''
        m0 = self.spectral_data.moments[0]
        Z1 = self.coeff[0]
        R = self.coeff[4]
        Q = self.coeff[7]
        G1 = self.coeff[3]
        G2 = self.coeff[5]
        G3 = self.coeff[6]

        def pdf(s):
            Z =  Z1*s
            px = (1.0/np.sqrt(m0)) * ( \
                            (G1/Q)*np.exp(-Z/Q) + \
                            ((G2*Z)/(R**2))*np.exp(-((Z)**2)/(2.*R**2)) + \
                            (G3*Z)*np.exp(-((Z)**2)/2.) \
                            )
            return px
        return pdf(s)

    def get_life(self, C, k, integrate_pdf=False):
        """Calculate fatigue life with parameters C, k, as defined in [2].

        :param C: [int,float]
            Fatigue strength coefficient [MPa**k].
        :param k : [int,float]
            Fatigue strength exponent [/].
        :param integrate_pdf:  boolean
            If true the the fatigue life is estimated by integrating the PDF, 
            Default is false which means that the theoretical equation is used
        :return T: float
            Estimated fatigue life in seconds.
        """ 
        if integrate_pdf:
            d = self.spectral_data.m_p / C * \
                quad(lambda s: s**k*self.get_PDF(s), 
                     a=0, b=np.Inf)[0]
        else:
            m0 = self.spectral_data.moments[0]
            
            R = self.coeff[4]
            Q = self.coeff[7]
            G1 = self.coeff[3]
            G2 = self.coeff[5]
            G3 = self.coeff[6]
            
            d = 1 / C * ( self.spectral_data.m_p 
                          * np.sqrt(m0)**k 
                          * (
                                G1 * (Q**k)*ss.gamma(1.0+k)+(np.sqrt(2.0)**k)*\
                                ss.gamma(1.+k/2.)*(G2 * abs(R)**k+G3)\
                            )
                        )
    
        T = 1/d
        return T


import numpy as np
from scipy.special import gamma
from scipy.integrate import quad

class Narrowband(object):
    """Class for fatigue life estimation using frequency domain 
    method by Miles [1].
   
    References
    ----------
    [1] John W. Miles. On structural fatigue under random loading. Journal
        of the Aeronautical Sciences, 21(11):753{762, 1954.
    [2] Janko Slavič, Matjaž Mršnik, Martin Česnik, Jaka Javh, Miha Boltežar. 
        Vibration Fatigue by Spectral Methods, From Structural Dynamics to Fatigue Damage
        – Theory and Experiments, ISBN: 9780128221907, Elsevier, 1st September 2020
    """
    def __init__(self, spectral_data):
        '''Get needed values from reference object.
        
        :param spectral_data:  Instance of object SpectralData
        '''                        
        self.spectral_data = spectral_data

    def get_PDF(self,s):
        '''Returns cycle PDF(Probability Density Function) as a function of stress s.

        :param s:  numpy.ndarray
            Stress vector.
        :return pdf: function pdf(s)
        '''
        m0 = self.spectral_data.moments[0]

        def pdf(s):
            px = (s / m0) * np.exp( - s**2 / (2 * m0)) 
            return px
        return pdf(s)

    def damage_intesity_NB(self, m0, nu, C, k):
        """Calculates narrowband damage intensity with parameters m0, nu, C, k, as defined in [2].

        :param m0: [int,float]
            Zeroth spectral moment [MPa**2].
        :param nu: [int,float]
            Frequency of positive slope zero crossing [Hz].
        :param C: [int,float]
            Fatigue strength coefficient [MPa**k].
        :param k : [int,float]
            Fatigue strength exponent [/].
        :return T: float
            Estimated fatigue life in seconds.       
        """
        d = nu * np.sqrt(2 * m0)**k * gamma(1.0 + k/2.0) / C  
        return  d

    def get_life(self, C, k, integrate_pdf=False):
        """Calculate fatigue life with parameters C, k, as defined in [2].

        :param C: [int,float]
            Fatigue strength coefficient [MPa**k].
        :param k : [int,float]
            Fatigue strength exponent [/].
        :param integrate_pdf:  boolean
            If true the the fatigue life is estimated by integrating the PDF, 
            Default is false which means that the theoretical equation is used
        :return T: float
            Estimated fatigue life in seconds.
        """ 
        if integrate_pdf:
            d = self.spectral_data.nu / C * \
                quad(lambda s: s**k*self.get_PDF(s), 
                     a=0, b=np.Inf)[0]
        else:
            m0 = self.spectral_data.moments[0]
            nu = self.spectral_data.nu
            d = self.damage_intesity_NB(m0=m0, nu=nu, C=C, k=k) 

        T = 1.0/d
        return T


from .dirlik import Dirlik
from .petrucci_zuccarello import PetrucciZuccarello
from .tovo_benasciutti import TovoBenasciutti
from .zhao_baker import ZhaoBaker
from .narrowband import Narrowband
from .alpha075 import Alpha075
from .wirsching_light import WirschingLight
from .rice import Rice
from .gao_moan import GaoMoan



import numpy as np
from scipy.special import gamma
from .narrowband import Narrowband

class Alpha075(Narrowband):
    """Class for fatigue life estimation using frequency domain 
    method by Benasciutti and Tovo [1]. 

    References
    ----------
    [1] Denis Benasciutti and Robert Tovo. Rainflow cycle distribution and
        fatigue damage in Gaussian random loadings. Technical report, Department
        of Engineering, University of Ferrara, 2004
    [2] Janko Slavič, Matjaž Mršnik, Martin Česnik, Jaka Javh, Miha Boltežar. 
        Vibration Fatigue by Spectral Methods, From Structural Dynamics to Fatigue Damage
        – Theory and Experiments, ISBN: 9780128221907, Elsevier, 1st September 2020
    """
    def __init__(self, spectral_data):
        '''Get needed values from reference object.

        :param spectral_data:  Instance of object SpectralData
        '''                        
        Narrowband.__init__(self, spectral_data)
    
    def get_life(self, C, k):
        """Calculate fatigue life with parameters C, k, as defined in [2].

        :param C: [int,float]
            Fatigue strength coefficient [MPa**k].
        :param k : [int,float]
            Fatigue strength exponent [/].
        :return T: float
            Estimated fatigue life in seconds.
        """
        m0 = self.spectral_data.moments[0]
        nu = self.spectral_data.nu
        a075 = self.spectral_data.m075 / np.sqrt(m0 * self.spectral_data.m150)

        dNB = self.damage_intesity_NB(m0=m0, nu=nu, C=C, k=k) 
        d =  a075**2 * dNB
        T = 1.0 / d
        
        return T


import numpy as np
import scipy.integrate as si
import scipy.stats as ss
from scipy.integrate import quad

class Rice(object):
    """Class for fatigue life estimation using frequency domain 
    method by Rice[1].

    References
    ----------
    [1] Stephen O. Rice. Mathematical analysis of random noise. The Bell
        System Technical Journal, 24(1):46-156, 1945.
    [2] Janko Slavič, Matjaž Mršnik, Martin Česnik, Jaka Javh, Miha Boltežar. 
        Vibration Fatigue by Spectral Methods, From Structural Dynamics to Fatigue Damage
        – Theory and Experiments, ISBN: 9780128221907, Elsevier, 1st September 2020
    """

    def __init__(self, spectral_data):
        '''Get needed values from reference object.
        '''                
        self.spectral_data = spectral_data


    def get_PDF(self, s):
        '''Returns cycle PDF(Probability Density Function) as a function of stress s.

        :param s:  numpy.ndarray
            Stress vector.
        :return pdf: function pdf(s)
        '''
        m0 = self.spectral_data.moments[0]
        al2 = self.spectral_data.al2

        def pdf(s):
            px = np.sqrt(1.0 - al2**2)/np.sqrt(2.0 * np.pi * m0) * \
                np.exp( - (s**2) / (2.0 * m0 * (1.0 - al2**2))) +\
                al2*s/m0 * np.exp( - (s**2) / (2*m0)) * \
                ss.norm.cdf((al2 * s) / (np.sqrt(m0 * (1 - al2**2))))
            return px
        return pdf(s)

    def get_life(self, C, k):
        """Calculate fatigue life with parameters C, k, as defined in [2].

        :param C: [int,float]
            Fatigue strength coefficient [MPa**k].
        :param k : [int,float]
            Fatigue strength exponent [/].
        :return T: float
            Estimated fatigue life in seconds.
        """ 
        d = self.spectral_data.m_p / C * \
            quad(lambda s: s**k*self.get_PDF(s), 
                    a=0, b=np.Inf)[0]

        T = 1.0/d

        return T


import numpy as np
from scipy.integrate import quad
from scipy.special import gamma
from scipy.optimize import fsolve

class ZhaoBaker(object):
    """Class for fatigue life estimation using frequency domain 
    method by Zhao and Baker[1, 2].
     
    References
    ----------
    [1] Wangwen Zhao and Michael J. Baker. On the probability density function
        of rainflow stress range for stationary Gaussian processes. International
        Journal of Fatigue, 14(2):121-135, 1992
    [2] Janko Slavič, Matjaž Mršnik, Martin Česnik, Jaka Javh, Miha Boltežar. 
        Vibration Fatigue by Spectral Methods, From Structural Dynamics to Fatigue Damage
        – Theory and Experiments, ISBN: 9780128221907, Elsevier, 1st September 2020
    """
    def __init__(self, spectral_data):
        '''Get needed values from reference object.

        :param spectral_data:  Instance of object SpectralData
        '''     
        self.spectral_data = spectral_data

    def _calculate_coefficients(self, method='method 1'):
        """Calculate coefficients for Zhao-Baker method.
        
        :param method:  string
            - 'method 1' is tuned in simulations with material parameters 
               in the range of 2 <= k <= 6, where k is S-N curve coefficient.
            - 'method 2' is derived for S-N curve coefficient k = 3.  
        :return [a, b, w]: list
            a and b are Weibull distribution coefficients. w is weight coefficient.
        """
        if method == 'method 1': 
            a, b, w = self._calculate_coefficients_method_1()
        elif method == 'method 2': 
            a, b, w = self._calculate_coefficients_method_2()
        else: raise Exception('Unrecognized Input Error')
        return a, b, w


    def _calculate_coefficients_method_1(self):
        """Calculate coefficients for Zhao-Baker method 1.
        
        Method 1 is tuned in simulations with material parameters 
        in the range of 2 <= k <= 6, where k is S-N curve coefficient.
        
        :return [a, b, w]: list
            a and b are Weibull distribution coefficients. w is weight coefficient.
        """
        al2 = self.spectral_data.al2
        
        a = 8.0 - 7.0 * al2 
        if al2 < 0.9:
            b = 1.1 
        else:
            b = 1.1 + 9.0 * (al2 - 0.9) 
        w = ( 1.0 - al2 ) / ( 1.0 - np.sqrt(2.0/np.pi) * gamma(1.0 + 1.0/b) * a**(-1.0/b) ) 
        
        return [a, b, w]

    def _calculate_coefficients_method_2(self):
        """Calculate coefficients for Zhao-Baker method 2.
        
        Method 2 is derived for S-N curve coefficient k = 3.  
        
        :return [a, b, w]: list
            a and b are Weibull distribution coefficients. w is weight coefficient.
        """
        al2 = self.spectral_data.al2
        a075 = self.spectral_data.al075
        
        if al2 < 0.9:
            b = 1.1 
        else:
            b = 1.1 + 9 * (al2 - 0.9) 
        
        if a075 >= 0.5:
            ro = -0.4154 + 1.392 * a075  #damage correction factor
        else:
            ro = 0.28  #damage correction factor
                
        def eq(p):
            return gamma(1.0+(3.0/b)) * (1.0-al2) * p**3.0 + \
                   3.0 * gamma(1.0+(1.0/b)) * (ro * al2 - 1.0) * p + \
                   3.0 * np.sqrt(np.pi/2.0) * al2 * (1.0 - ro)
        
        try:
            root = fsolve(eq, 0)
        except:
            root = fsolve(eq, np.random.rand()*5.0)
        
        a = root**(-b) 
        w = ( 1.0 - al2 ) / ( 1.0 - np.sqrt(2.0/np.pi) * gamma(1.0 + 1.0/b) * a**(-1.0/b) )

        return [a, b, w]

    def get_PDF(self, s, method='method 2'):
        '''Returns cycle PDF(Probability Density Function) as a function of stress s.

        :param s:  numpy.ndarray
            Stress vector.
        :param method:  string
            - 'method 1' is tuned in simulations with material parameters 
               in the range of 2 <= k <= 6, where k is S-N curve coefficient.
            - 'method 2' is derived for S-N curve coefficient k = 3. 
        :return pdf: function pdf(s)
        '''
        m0 = self.spectral_data.moments[0]
        a, b, w = self._calculate_coefficients(method=method)

        def pdf(s):
            return  w * ((a*b) / (np.sqrt(m0))) * ((s/np.sqrt(m0)))**(b-1) * np.exp(-a * (s/np.sqrt(m0))**b) +\
                (1-w) * (s/m0) * np.exp(-0.5 * (s/np.sqrt(m0))**2)
        return pdf(s)
        
        
    def get_life(self, C, k, method='method 1', integrate_pdf=False): 
        """Calculate fatigue life with parameters C and k, as defined in [2].

        :param C: [int,float]
            Fatigue strength coefficient [MPa**k].
        :param k : [int,float]
            Fatigue strength exponent [/].
        :param method : str
            - 'method 1' is tuned in simulations with material parameters 
               in the range of 2 <= k <= 6, where k is S-N curve coefficient.
            - 'method 2' is derived for S-N curve coefficient k = 3.  
        :param integrate_pdf:  boolean
            If true the the fatigue life is estimated by integrating the PDF, 
            Default is false which means that the theoretical equation is used
        :return T: float
            Estimated fatigue life in seconds.
        """
        if integrate_pdf:
            d = self.spectral_data.m_p / C * \
                quad(lambda s: s**k*self.get_PDF(s, method=method), 
                    a=0, b=np.Inf)[0]
        else:
            m0 = self.spectral_data.moments[0]
            m_p = self.spectral_data.m_p

            a, b, w = self._calculate_coefficients(method=method)
                    
            d = (m_p/C) * m0**(0.5*k) * ( w * a**(-k/b) * gamma(1.0+k/b) +\
                    (1.0-w) * 2**(0.5*k) * gamma(1.0+0.5*k) )
    
        T = float(1.0/d)
        return T


import numpy as np
from scipy.special import gamma
from .narrowband import Narrowband

class WirschingLight(Narrowband):
    """Class for fatigue life estimation using frequency domain 
    method by Wirsching and Light [1].
   
    References
    ----------
    [1] Paul H. Wirsching and Mark C. Light. Fatigue under wide band random
        stresses. Journal of the Structural Division, 106(7):1593-1607,
        1980
    [2] Janko Slavič, Matjaž Mršnik, Martin Česnik, Jaka Javh, Miha Boltežar. 
        Vibration Fatigue by Spectral Methods, From Structural Dynamics to Fatigue Damage
        – Theory and Experiments, ISBN: 9780128221907, Elsevier, 1st September 2020
    """
    def __init__(self, spectral_data):
        '''Get needed values from reference object.

        :param spectral_data:  Instance of object SpectralData
        '''
        Narrowband.__init__(self, spectral_data)
                           
    def get_life(self, C, k):
        """Calculate fatigue life with parameters C, k, as defined in [2].

        :param C: [int,float]
            Fatigue strength coefficient [MPa**k].
        :param k : [int,float]
            Fatigue strength exponent [/].
        :return T: float
            Estimated fatigue life in seconds.
        """
        m0 = self.spectral_data.moments[0]
        nu = self.spectral_data.nu
        al2 = self.spectral_data.al2

        dNB = self.damage_intesity_NB(m0=m0, nu=nu, C=C, k=k) 
        ak = 0.926 - 0.033 * k
        bk = 1.587 * k - 2.323
        eps = np.sqrt(1 - al2**2)

        ro = ak + ( 1 - ak ) * ( 1 - eps )**bk
        T = 1 / (ro * dNB)
        
        return T


import numpy as np
import fatpack

class Rainflow(object):
    """Class for fatigue life estimation using rainflow counting method [1].

    References
    ----------
    [1] C. Amzallag et. al. Standardization of the rainflow counting method for
        fatigue analysis. International Journal of Fatigue, 16 (1994) 287-293
    [2] Janko Slavič, Matjaž Mršnik, Martin Česnik, Jaka Javh, Miha Boltežar. 
        Vibration Fatigue by Spectral Methods, From Structural Dynamics to Fatigue Damage
        – Theory and Experiments, ISBN: 9780128221907, Elsevier, 1st September 2020
    """
    def __init__(self, spectral_data):
        '''Get needed values from reference object.

        :param spectral_data:  Instance of object SpectralData
        '''          
        self.spectral_data = spectral_data


    def get_life(self, C, k, nr_load_classes = 512, Su = False, range = False):
        """Calculate fatigue life with parameters C, k, as defined in [2].
 
        :param C: [int,float]
            Fatigue strength coefficient [MPa**k].
        :param k : [int,float]
            Fatigue strength exponent [/].
        :param nr_load_classes: int
            The number of intervals to divide the min-max range of the dataseries
            into. Defaults to 512.
        :param Su: [int,float]
            Ultimate tensile strength [MPa]. If specified, Goodman equivalent stress
            is used for fatigue life estimation. Defaults to False.
        :param range: bool
            If True, ranges instead of amplitudes are used for fatigue life estimation.
            Defaults to False.
        :return T: float
            Estimated fatigue life in seconds.
        """    
        t = self.spectral_data.t
        ranges,means = fatpack.find_rainflow_ranges(self.spectral_data.data, k=nr_load_classes, return_means=True)
        
        if range == True: pass
        elif range == False: ranges *= 0.5
        else: raise Exception('Unrecognized Input Error')

        if Su: ranges = ranges/(1. - means/Su)

        d = np.sum(ranges**k / C)
        T = t / d
        
        return T



def relative_error(value, value_ref):
    """Return the relative error 

    :param value: [int,float]
        Value to compare
    :param value: [int,float]
        Reference value
    :return err: float
        Relative error.
    """ 
    return (value-value_ref)/value_ref

def basquin_to_sn(Sf, b, range = False):
    """Converts Basquin equation parameters Sf and b to fatigue life parameters C and k,
    as defined in [2]. Basic form of Basquin equation is used here: Sa = Sf* (2*N)**k 

    :param Sf: [int,float]
        Fatigue strength coefficient [MPa**k].
    :param b : [int,float]
        Fatigue strength exponent [/].
    :param range : bool
        False/True sets returned value C with regards to amplitude / range count, respectively.
    :return C, k: float
        C - fatigue strength coefficient [MPa**k], k - fatigue strength exponent [/].
    """ 
    if not range:
        k = -1/b 
        C = 0.5*Sf**k
        return C,k

    else: 
        k = -1/b 
        C = 0.5*(2*Sf)**k
        return C,k


import numpy as np
import scipy.stats as ss
import scipy.integrate as si
import scipy.special as spe
import warnings
from .narrowband import Narrowband

class GaoMoan(Narrowband):
    """Class for fatigue life estimation using frequency domain 
    method by Gao and Moan [1].
    
    :param spectral_data:  Instance of object SpectralData       
    :param band_frequency:  list
        List of frequencies that divides general wideband procces into 3 or less narrowband
        processes. Specified frequency is considered as upper limit of narrowband process.
        Defaults to False.

    References
    ----------
    [1] Zhen Gao and Torgeir Moan. Frequency-domain fatigue analysis of
        wide-band stationary Gaussian processes using a trimodal spectral formulation.
        International Journal of Fatigue, 30(10-11):1944{1955, 2008
    [2] Janko Slavič, Matjaž Mršnik, Martin Česnik, Jaka Javh, Miha Boltežar. 
        Vibration Fatigue by Spectral Methods, From Structural Dynamics to Fatigue Damage
        – Theory and Experiments, ISBN: 9780128221907, Elsevier, 1st September 2020
    """
    def __init__(self, spectral_data, band_frequency=False):
        '''Get needed values from reference object.

        :param spectral_data:  Instance of object SpectralData       
        :param band_frequency:  list
            List of frequencies that divides general wideband procces into 3 or less narrowband
            processes. Specified frequency is considered as upper limit of narrowband process.
            Defaults to False.
         '''
        Narrowband.__init__(self, spectral_data)

        band = [] 
        if isinstance(band_frequency, list):
            for indx,freq in enumerate(band_frequency):
                if indx <3:
                    if isinstance(freq, (int,float)):
                        band.append(freq)
                    else:
                        raise Exception('Unrecognized Input Error')
                else:
                    warnings.warn("Too many frequency bands provided. Onyl first 3 will be used.")
                    break

        self.band_frequency = np.array(band)
        self.bands = False 
        
    def get_life(self, C, k):
        """Calculate fatigue life with parameters C, k, as defined in [2].

        :param C: [int,float]
            Fatigue strength coefficient [MPa**k].
        :param k : [int,float]
            Fatigue strength exponent [/].
        :return T: float
            Estimated fatigue life in seconds.
        """
        
        if len(self.band_frequency) == 1: # narrow-band, user specified bandwidth

            psd = self.spectral_data.psd
            f1 = self.band_frequency[0]     
            f1_indx = np.abs(psd[:, 0]  - f1).argmin()

            # -- spectral moments, positive slope zero crossing frequency
            m0H,_,m2H,_,_ = self.spectral_data.calculate_spectral_moments_frequency(psd[:f1_indx, :])
            v0H = np.sqrt(m2H / m0H)

            # -- Define expected value of stress range ( int(S^k * p(s)) ) proces H(t), fatigue life
            dNB_H = self.damage_intesity_NB(m0=m0H, nu=v0H, C=C, k=k) 
            T = 1/dNB_H
        
            return T
            
        elif len(self.band_frequency) == 2: # bi-modal, user specified bandwidth
            
            psd = self.spectral_data.psd
            f1, f2 = self.band_frequency[0], self.band_frequency[1] 
            f1_indx = np.abs(psd[:, 0] - f1).argmin()
            f2_indx = np.abs(psd[:, 0] - f2).argmin()
            
            # -- spectral moments
            m0H,m1H,m2H,_,_ = self.spectral_data.calculate_spectral_moments_frequency(psd[:f1_indx,:])
            m0M,_,m2M,_,_ = self.spectral_data.calculate_spectral_moments_frequency(psd[f1_indx:f2_indx,:])
            
            # -- Vanmarcke bandwidth parameter
            eps_H = np.sqrt(1.0 - m1H**2/(m0H*m2H))
            
            # -- positive slope zero crossing frequency
            v0H = np.sqrt(m2H / m0H)
            v0P = np.sqrt(m2H * eps_H**2 + m2M) * (np.sqrt(m0M) /  (m0H + m0M))
            
            # -- peak pdf 
            pdf_H = lambda x: ss.rayleigh.pdf(x / np.sqrt(m0H)) / np.sqrt(m0H)
            pdf_M = lambda x: ss.rayleigh.pdf(x / np.sqrt(m0M)) / np.sqrt(m0M)
        
            # -- max stress is 3*standard deviation
            smax = 3.0 * np.sqrt(self.spectral_data.moments[0])
            dx = 0.05
            x = np.arange(0, smax, dx) 
            
            pdf_P = dx * np.convolve(pdf_H(x), pdf_M(x)) 
            x2 = np.arange(0, pdf_P.size * dx, dx)
            
            # -- Define expected value of stress range ( int(S^k * p(s)) ) P(t)
            SP = si.simps(x2**k * pdf_P, x2, dx)

            dNB_H = self.damage_intesity_NB(m0=m0H, nu=v0H, C=C, k=k) 
            dNB_P = v0P * SP / C
            
            d = dNB_H + dNB_P
            T = 1 / d
                
            return T
            
        elif len(self.band_frequency) == 3: # tri-modal, user specified band width
            self.bands = True 
            
            psd = self.spectral_data.psd
            f1, f2, f3 = self.band_frequency[0], self.band_frequency[1], self.band_frequency[2] 
            f1_indx = np.abs(psd[:, 0] - f1).argmin()
            f2_indx = np.abs(psd[:, 0] - f2).argmin()
            f3_indx = np.abs(psd[:, 0] - f3).argmin()

            self.frequencies_indx = (f1_indx, f2_indx, f3_indx)

            return self._calculate_life_trimodal(C, k)  
            
        else: # tri-modal, equal variance bandwidth
            return self._calculate_life_trimodal(C, k)  

    def _calculate_life_trimodal(self, C, k):
        '''Returns fatique life based on specified 3-modal process bandwith.
        '''
        # -- calculate equal area bands and its moments
        m0H, m1H, m2H, m0M, m1M, m2M, m0L, _, m2L = self.getBands()

        # -- positive slope zero crossing frequency
        v0H = np.sqrt(m2H/m0H)

        # -- Vanmarcke bandwidth parameter
        eps_H, eps_M = [np.sqrt(1.0 - m1H**2/(m0H*m2H)), np.sqrt(1.0 - m1M**2/(m0M*m2M))]
        
        # -- positive slope zero crossing frequency
        v0P = np.sqrt(m2H * eps_H**2 + m2M) * (np.sqrt(m0M) / (m0H + m0M))
        v0Q = np.sqrt(m2H * eps_H**2 + m2M * eps_M**2 + m2L) * \
            (2.0 * np.sqrt(m0L * (m0H + m0M + m0L)) - np.pi * np.sqrt(m0H * m0M) \
            + 2.0 * np.sqrt(m0H * m0M) * np.arctan(np.sqrt((m0H * m0M ) /m0L) / np.sqrt(m0H + m0M + m0L))) \
            / (2.0 * np.sqrt(m0H + m0M + m0L)**3)

        # -- peak pdf 
        pdf_H = lambda x: ss.rayleigh.pdf(x/np.sqrt(m0H)) / np.sqrt(m0H)
        pdf_M = lambda x: ss.rayleigh.pdf(x/np.sqrt(m0M)) / np.sqrt(m0M)
        pdf_L = lambda x: ss.rayleigh.pdf(x/np.sqrt(m0L)) / np.sqrt(m0L)
        
        # -- max stress is 3*standard deviation
        smax = 3.0 * np.sqrt(self.spectral_data.moments[0])
        dx = 0.05
        x = np.arange(0, smax, dx)
        
        pdf_P = dx * np.convolve(pdf_H(x), pdf_M(x))
        x2 = np.arange(0, pdf_P.size * dx, dx)
        pdf_Q = dx * np.convolve(pdf_P, pdf_L(x2))
        x3 = np.arange(0, pdf_Q.size*dx , dx)
        
        # -- Define expected value of stress range ( int(S^k * p(s)) )
        SP = si.simps(x2**k * pdf_P, x2, dx)  #quad funkcija
        SQ = si.simps(x3**k * pdf_Q, x3, dx)

        # -- Calculate damage intensity d, fatigue life
        dNB_H = self.damage_intesity_NB(m0=m0H, nu=v0H, C=C, k=k) 
        dNB_P = v0P * SP / C
        dNB_Q = v0Q * SQ / C
        
        d = dNB_H + dNB_P + dNB_Q
        T = 1/ d
        
        return T

    def getBands(self):  
        '''Produce data for calculation
        based on three frequency bands'''
        
        if not self.bands:
            frq = self.equalAreaBands()
        else:
            frq = self.frequencies_indx
        
        psd = self.spectral_data.psd

        m_L = self.spectral_data.calculate_spectral_moments_frequency(psd[:frq[0],:])[:3] #only fist 3 moments
        m_M = self.spectral_data.calculate_spectral_moments_frequency(psd[frq[0]:frq[1],:])[:3]
        m_H = self.spectral_data.calculate_spectral_moments_frequency(psd[frq[1]:frq[2],:])[:3]
        
        return np.hstack((m_H, m_M, m_L))

    
    def equalAreaBands(self):
        '''Divide PSD in three bands with
        equal area. Bands border at frequncys f1, f2 and f3.
        '''
        psd = self.spectral_data.psd[:,1]
        q_area = np.sum(psd) # (not the REAL area because x-axis unit is omitted)
        
        # -- Calculate a cumulative sum vector of psd
        c_psd = np.cumsum(psd)
        
        # -- Find where cumulative sum is equal to 1/3, 2/3 and 3/3 of A
        f1 = np.abs(c_psd - q_area/3.0).argmin()
        f2 = np.abs(c_psd - 2.0 * q_area/3.0).argmin()
        f3 = np.abs(c_psd - 3.0 * q_area/3.0).argmin()
         
        return (f1, f2, f3)


import numpy as np

class PetrucciZuccarello(object):
    """Class for fatigue life estimation using frequency domain 
    method by Petrucci and Zuccarello [1].

    References
    ----------
    [1] G. Petrucci and B. Zuccarello. Fatigue life prediction under wide band
        random loading. Fatigue & Fract. Eng. Mater. & Struct., 27(12):1183-1195, 
        December 2004.
    [2] Janko Slavič, Matjaž Mršnik, Martin Česnik, Jaka Javh, Miha Boltežar. 
        Vibration Fatigue by Spectral Methods, From Structural Dynamics to Fatigue Damage
        – Theory and Experiments, ISBN: 9780128221907, Elsevier, 1st September 2020
    """
    def __init__(self, spectral_data):
        '''Get needed values from reference object.

        :param spectral_data:  Instance of object SpectralData
        '''          
        self.spectral_data = spectral_data
    
    def get_life(self, C, k, Su = 1110):
        """Calculate fatigue life with parameters C, k, as defined in [2].

        :param C: [int,float]
            Fatigue strength coefficient [MPa**k].
        :param k : [int,float]
            Fatigue strength exponent [/].
        :param Su: [int,float]
            Tensile strength [MPa].    
        :return Tf: float
            Estimated fatigue life in seconds.
        """
        m0,m1,m2,_,m4 = self.spectral_data.moments
        m_p = self.spectral_data.m_p
        
        a1, a2 = (  np.sqrt((m1**2)/(m0*m2)) ,  m2 / np.sqrt(m0*m4) )
        R = 3 * np.sqrt(m0) / Su 

        alpha_m = np.array([
                        [-1, -a2, a1, a1*a2, -a2**2, -a1**2],
                        [ 1, -a2, a1, a1*a2, a2**2, -a1**2],
                        [-1, -a2, a1, a1*a2, a2**2, -a1**2],
                        [ 1, -a2, a1, a1*a2, a2**2, -a1**2]
                        ])
        
        psi_m = np.array([
                        [1.994,  9.381, 18.349, 15.261, 1.483, 15.402],
                        [8.229, 26.510, 21.522 ,27.748, 4.338, 20.026],
                        [0.946,  8.025, 15.692, 11.867, 0.382, 13.198],
                        [8.780, 26.058, 21.628, 26.487, 5.379, 19.967]
                        ])
        
        p1 = np.dot(alpha_m[0], psi_m[0])
        p2 = np.dot(alpha_m[1], psi_m[1])
        p3 = np.dot(alpha_m[2], psi_m[2])
        p4 = np.dot(alpha_m[3], psi_m[3])
        
        psi = ( (p2-p1)/6.0 ) * (k-3.0) + p1 + ( (2.0/9.0)*(p4-p3-p2+p1)*(k-3) + \
                                                (4.0/3.0)*(p3-p1) ) * (R - 0.15)

        Tf = C / ( m_p * np.sqrt( m0**k ) * np.exp(psi) )

        return Tf